"""
downloader.py

this module contains the functionality for downloading downloading
many urls concurrently with a worker queue+semaphore asyncio approach
"""
from __future__ import annotations
import os
import re
from pathlib import Path
import asyncio
import shutil
from time import perf_counter
from typing import Iterable, Set, Tuple
from dataclasses import dataclass
from tqdm import tqdm
from aiohttp import ClientSession, ClientTimeout, TCPConnector, ClientResponseError, InvalidURL
@dataclass
class _FormsiteDownloader:

    """Handles the distribution of downlaod tasks across DownloadWorkers."""
    download_folder: str
    links: Iterable[str]
    max_concurrent_downloads: int = 10
    timeout: int = 80
    retries: int = 1
    filename_regex: str = r''
    strip_prefix: bool = False
    overwrite_existing: bool = True
    report_downloads: bool = False
    write_failed: bool = True
    display_progress: bool = True

    def __post_init__(self):
        """Initializes internal variables."""
        self.filename_compiled_regex = re.compile(self.filename_regex)
        self.semaphore = asyncio.Semaphore(self.max_concurrent_downloads)
        self.dl_queue = asyncio.Queue()
        if not self.overwrite_existing and len(self.links) > 0:
            for element in self.links:
                url = element.rsplit('/', 1)[0]+"/"
                break
            filenames_in_dl_dir = self._list_files_in_download_dir(url)
            self.links = set(self.links) - filenames_in_dl_dir
        else:
            self.links = set(self.links)
        self.internal_state = DownloadWorkerState(self.links, self.max_concurrent_downloads)

    async def Start(self) -> None:
        """Starts download of links."""
        pbar = tqdm(total=len(self.links),
                    desc="Downloading files",
                    unit='f',
                    leave=False,
                    dynamic_ncols=True,
                    ncols=80) if self.display_progress else None
        os.makedirs(self.download_folder, exist_ok=True)
        async with ClientSession(connector=TCPConnector(limit=0)) as session:
            self.internal_state.update_pbar_callback = pbar.update if pbar is not None else None
            for link in self.links:
                self.dl_queue.put_nowait((link, 0))
            tasks = [asyncio.ensure_future(
                DownloadWorker(self.download_folder,
                               self.dl_queue,
                               self.semaphore,
                               session,
                               self.internal_state,
                               timeout=self.timeout,
                               retries=self.retries,
                               pbar=pbar,
                               filename_regex=self.filename_regex,
                               strip_prefix=self.strip_prefix).main())
                     for _ in range(self.max_concurrent_downloads)]
            await asyncio.gather(*tasks)
        pbar.close()
        if self.internal_state.failed > 0 and self.write_failed:
            self.internal_state.write_failed()
        if self.report_downloads:
            self.internal_state.write_all()

    def _list_files_in_download_dir(self, url: str) -> Set[str]:
        """Lists all files in `self.download_folder`, inserts `url` before the filename."""
        filenames_in_dir = set()
        for file in os.listdir(self.download_folder):
            filenames_in_dir.add(url+file)
        return filenames_in_dir

@dataclass
class DownloadWorkerState:

    """Contains methods that track internal downlaod progressa cross downloader workers."""
    urls: Iterable[str]
    active_workers: int

    def __post_init__(self):
        self.urls = set(self.urls)
        self.total: int = len(self.urls)
        self.enqueued: int = len(self.urls)
        self.in_progress: int = 0
        self.success: int = 0
        self.failed: int = 0
        self.failed_urls: set = set()
        self.success_urls: set = set()
        self.complete_urls: list = list()
        self.update_pbar_callback = None
        self.last_progress_display_update = 0

    def total_complete(self):
        """Returns sum of `self.success` and `self.failed`"""
        return self.success + self.failed

    def write_failed(self, target: str = './failed_downloads.txt'):
        """Writes contents of `self.failed_urls` to a file."""
        with open(target, 'a') as writer:
            to_write = self.get_failed_url_diff()
            for i in to_write:
                writer.write(f"{i}\n")
        print(f"{len(to_write)} failures, please see failed_downloads.txt for more info"
              "\nfor timeout errors, try increasing timeout or reducing max concurrent downloads")

    def write_all(self, target: str = './downloads_status.txt'):
        """Writes contents of `self.complete_urls` to a file."""
        with open(target, 'w') as writer:
            for i in self.complete_urls:
                writer.write(f"{i}\n")

    def display_relevant_info(self):
        """Debug: prints some relevant internal state values"""
        print('----------------------------------')
        print(f"enqueued: {self.enqueued}")
        print(f"active workers: {self.active_workers}")
        print(f"files in progress: {self.in_progress}")
        print(f"success: {self.success}")
        print(f"failed: {self.failed}")

    def try_exit(self) -> bool:
        """Checks if it is possible to terminate worker."""
        if self.total_complete() >= self.total:
            return True
        elif self.enqueued > 0:
            return False
        elif self.active_workers > self.in_progress:
            return True
        else: return False

    def can_terminate_worker(self) -> bool:
        """Decrements allowed number of active workers if conditions allow it."""
        if self.try_exit():
            self.active_workers -= 1
            return True
        else: return False

    def mark_success(self, url: str) -> None:
        """Increments internal counter to match completed downloads."""
        self._mark(url, "OK", self.success_urls, self.success)

    def mark_fail(self, url: str, fail_exception: Exception) -> None:
        """Increments internal counter to match completed downloads."""
        self._mark(url, repr(fail_exception), self.failed_urls, self.failed)

    def _mark(self, url: str, status: str, add_to: Set[str], count_to: int):
        """Base method for changing internal state counts."""
        self.complete_urls.append(f"{url};\t{status}")
        add_to.add(f"{url}")
        count_to += 1
        if self.update_pbar_callback is not None:
            self.update_pbar_callback(1)

    def end_iteration(self) -> None:
        """Runs at the end of a download iteration."""
        self.in_progress -= 1

    def start_iteration(self) -> None:
        """Runs at the start of a download iteration."""
        self.enqueued -= 1
        self.in_progress += 1

    def get_failed_url_diff(self) -> Set[str]:
        """Returns a difference between all input URLs and successfully downloaded urls."""
        return set(self.urls) - set(self.success_urls)

@dataclass
class DownloadWorker:

    """download_folder is a path to download directory,
    \nqueue, semaphore, session, internal state and pbar are shared across all workers
    """
    download_folder: str
    queue: asyncio.Queue
    semaphore: asyncio.Semaphore
    session: ClientSession
    internal_state: DownloadWorkerState
    timeout: int = 80
    retries: int = 1
    pbar: tqdm = None
    filename_regex: str = r''
    strip_prefix: bool = False

    def __post_init__(self):
        self.client_timeout = ClientTimeout(total=self.timeout)
        self.filename_compiled_regex = re.compile(self.filename_regex)
        self.display_progress = None if self.pbar is None else True

    async def main(self) -> int:
        """Entrypoint for launching the download process."""
        while True:
            try:
                if self.internal_state.can_terminate_worker():
                    break
                url, attempt = await self.queue.get()
                await self.semaphore.acquire()
                if abs(self.internal_state.last_progress_display_update - perf_counter()) > 5: # seconds
                    self._update_pbar(desc="Downloading files")
                self.internal_state.start_iteration()
                try:
                    response = await self._download(url)
                    if response == 0:
                        self.internal_state.mark_success(url)
                except Exception as download_exception:
                    self._error_handling(download_exception, url, attempt)
            except asyncio.CancelledError:
                pass
            finally:
                self.semaphore.release()
                self.internal_state.end_iteration()
        return 0

    async def _fetch(self, url: str, filename: str, target: str, chunk_size: int = 4*1024,
                     in_progress_ext: str = '.tmp') -> int:
        """The core download function with `session.get` request."""
        display_name = filename[:20]+"…"+filename[-5:] if len(filename) > 25 else filename[:25]
        async with self.session.get(url, timeout=self.client_timeout) as response:
            response.raise_for_status()
            pbar = tqdm(desc=display_name,
                        total=response.content_length,
                        leave=False,
                        unit='b',
                        unit_scale=True,
                        unit_divisor=1024,
                        ncols=80) if self.display_progress else None
            with open(target+in_progress_ext, "wb") as f:
                async for chunk in response.content.iter_chunked(chunk_size):
                    f.write(chunk)
                    if pbar is not None:
                        pbar.update(len(chunk))
        try:
            pbar.close()
        except AttributeError:
            pass
        shutil.move(target+in_progress_ext, target)
        return 0

    async def _download(self, url: str) -> int:
        """Wrapper for starting the `self.fetch` operation."""
        filename, target = self.get_filename(url)
        exit_code = await self._fetch(url, filename, target)
        return exit_code

    def _error_handling(self, exception: Exception, url: str, attempt: int) -> None:
        """Decides what errors cancel downloads and which are retried."""
        if isinstance(exception, ClientResponseError):
            if exception.status == 403:
                self._update_pbar(desc=f"{exception}"[:79]+"…")
                self.internal_state.last_progress_display_update = perf_counter()
                self.internal_state.mark_fail(url, exception)
            elif exception.status == 404:
                self._update_pbar(desc=f"{exception}"[:79]+"…")
                self.internal_state.last_progress_display_update = perf_counter()
                self.internal_state.mark_fail(url, exception)
            else:
                self._retry_dl(url, attempt)
        elif isinstance(exception, InvalidURL):
            self.internal_state.mark_fail(url, exception)
        elif attempt < self.retries:
            self._retry_dl(url, attempt)
        else:
            self._update_pbar(desc=f"Failed {url}"[:79]+"…")
            self.internal_state.last_progress_display_update = perf_counter()
            self.internal_state.mark_fail(url, exception)

    def _retry_dl(self, url: str, attempt: int) -> None:
        """Puts failed downloads to the end of queue, if attempt < max retry."""
        self._update_pbar(desc=f"Retrying ({attempt}/{self.retries}) download of {url}"[:79]+"…")
        self.internal_state.last_progress_display_update = perf_counter()
        attempt += 1
        self.internal_state.enqueued += 1
        self.queue.put_nowait((url, attempt))

    def _update_pbar(self, desc: str = ""):
        """Updates main download progress bar with a provided description."""
        if isinstance(self.pbar, tqdm):
            self.pbar.set_description(desc=desc, refresh=True)

    def get_filename(self, url: str) -> Tuple[str, str]:
        """Gets filename from url. Returns filename and path+filename as target."""
        filename = f"{url.split('/')[-1:][0]}"
        if self.strip_prefix:
            filename = re.sub(r'^f-[\d]*-[\d]*-', r'', filename)
        if self.filename_compiled_regex.pattern != '':
            filename = self._regex_substitution(filename, self.filename_compiled_regex)
            target = f"{self.download_folder}/{filename}"
            target = self._check_if_file_exists(target)
        else:
            target = f"{self.download_folder}/{filename}"
        return filename, target

    def _check_if_file_exists(self, filename: str, appended_number: int = 0) -> str:
        """If `overwrite_downloads is False`, checks which URLs to omit based on filenames."""
        path = Path(filename).resolve()
        if path.exists():
            temp = filename.rsplit('.', 1)
            if temp[0].endswith(f'_{appended_number}'):
                temp[0] = temp[0][:temp[0].rfind(f'_{appended_number}')]
            try:
                filename = temp[0] + f"_{appended_number+1}." + temp[1]
            except IndexError:
                filename = temp[0] + f"_{appended_number+1}"
            filename = self._check_if_file_exists(filename, appended_number=appended_number+1)
        return filename

    def _regex_substitution(self, filename: str, filename_regex: re.Pattern) -> str:
        """Removes characters that match regex."""
        try:
            old_filename_tuple = filename.rsplit('.', 1)
            try:
                new_filename = f"{filename_regex.sub('', old_filename_tuple[0])}.{old_filename_tuple[1]}"
            except:
                new_filename = f"{filename_regex.sub('', old_filename_tuple[0])}"
        except:
            new_filename = f"{filename_regex.sub('', filename)}"
        return new_filename



"""
processing.py

This module handles processing of results from API jsons. Returns a dataframe.
"""
from __future__ import annotations
from datetime import datetime as dt
from typing import Any, Dict, Iterable
from dataclasses import dataclass
from tqdm import tqdm
import pandas as pd

@dataclass
class _FormsiteProcessing:

    """Handles processing of results from API jsons. Invoked with `self.Process()`"""
    items: dict
    results: Iterable[dict]
    interface: Any
    sort_asc: bool = False
    display_progress: bool = True

    def __post_init__(self):
        """Loads json strings as json objects."""
        self.timezone_offset = self.interface.params.timezone_offset
        self.column_map = self._generate_columns()
        self.metadata_map = self._generate_metadata()

    def _generate_columns(self) -> Dict[int,str]:
        """Creates a dict that maps column id to column label from items_json

        Returns:
            Dict[int,str]: Mapping of ID:Label
        """
        assert isinstance(self.items, dict), "items.json is empty"
        assert len(self.items) > 0, "items.json is empty"
        column_map = dict()
        for item in self.items['items']:
            column_map[int(item['id'])] = item['label']
        colmns_list = list(column_map.values())
        assert len(colmns_list) > 0, "Columns list is empty"
        return column_map

    def _generate_metadata(self) -> Dict[str,str]:
        """Creates a dict that maps metadata_label (hardcoded by formsite) to a friendlier export name

        Returns:
            Dict[str,str]: Mapping of metadata_ID: metadata_Label
        """
        metadata_map = {
            'id':'Reference #',
            'result_status': 'Status',
            'date_start': 'Start Time',
            'date_finish': 'Finish Time',
            'date_update': 'Date',
            'user_ip': 'User',
            'user_browser': 'Browser',
            'user_device': 'Device',
            'user_referrer': 'Referrer',
            'payment_status': 'Payment Status',
            'payment_amount': 'Payment Amount Paid',
            'login_username': 'Login Username',
            'login_email': 'Login Email',
        }
        return metadata_map       
    def _process_items_row(self, items: dict) -> pd.Series:
        """Converts a single 'items' record to a pandas Series

        Args:
            items (dict): ['items'] key from a results dict record

        Returns:
            pd.Series: a row of the new output csv
        """
        cols = []
        row = []
        for t in items:
            cols.append(int(t['id']))
            if t.get('values') is not None:
                values = []
                for val in t['values']:
                    values.append(val['value'])
                value = ' | '.join(values)
            else:
                value = t['value']
            row.append(value)
        return pd.Series(row, cols)

    def _process_metadata_row(self, result: dict) -> pd.Series:
        """Converts all non-['items'] keys to a pd.Series row

        Args:
            result (dict): an single results json record

        Returns:
            pd.Series: A row containing all metadata-related information
        """
        if 'items' in result:
            del result['items']
        return pd.Series(result)
    
    def _process_row(self, in_json: dict):
        """Merges metadata_row and items_row into a single row.

        Args:
            in_json (dict): a single results record json

        Returns:
            pd.Series: a row
        """
        items = self._process_items_row(in_json['items'])
        metadata = self._process_metadata_row(in_json)
        concated = pd.concat((metadata, items), axis=0)
        return concated

    def _reorder_columns(self, df: pd.DataFrame) -> pd.DataFrame:
        """Orders existing columns into a standard export format.
        First is always Reference #, Status, then Items, then Remaining metadata.

        Args:
            df (pd.DataFrame): Unordered DataFrame

        Returns:
            pd.DataFrame: Ordered DataFrame
        """
        # order:
        # Reference #, Status
        # items
        # Date,Start Time,Finish Time,Duration (s),User,Browser,Device,Referrer
        left_side = []
        right_side = []
        if 'id' in df.columns:
            left_side.append('id')
        if 'result_status' in df.columns:
            left_side.append('result_status')
        middle = [col if (col in df.columns) and (col not in self.metadata_map.keys()) else None for col in list(self.column_map.keys())]
        while None in middle:
            middle.remove(None)
        if 'payment_status' in df.columns:
            right_side.append('payment_status')
        if 'payment_amount' in df.columns:
            right_side.append('payment_amount')
        if 'login_username' in df.columns:
            right_side.append('login_username')
        if 'login_email' in df.columns:
            right_side.append('login_email')
        if 'date_update' in df.columns:
            right_side.append('date_update')
        if 'date_start' in df.columns:
            right_side.append('date_start')
        if 'date_finish' in df.columns:
            right_side.append('date_finish')
        if 'Duration (s)' in df.columns:
            right_side.append('Duration (s)')
        if 'user_ip' in df.columns:
            right_side.append('user_ip')
        if 'user_browser' in df.columns:
            right_side.append('user_browser')
        if 'user_device' in df.columns:
            right_side.append('user_device')
        if 'user_referrer' in df.columns:
            right_side.append('user_referrer')
        final = left_side+middle+right_side
        return df[final]

    def _cast_dtypes(self, df: pd.DataFrame) -> pd.DataFrame:
        """Casts the type of each existing column to their standard form.

        Args:
            df (pd.DataFrame): DataFrame with erratic dtypes

        Returns:
            pd.DataFrame: DataFrame with set dtypes
        """
        if 'Reference #' in df.columns:
            df['Reference #'] = df['Reference #'].astype(int)
        if 'Status' in df.columns:
            df['Status'] = df['Status'].astype(str)
        if 'Payment Status' in df.columns:
            df['Payment Status'] = df['Payment Status'].astype(str)
        if 'Payment Amount Paid' in df.columns:
            df['Payment Amount Paid'] = df['Payment Amount Paid'].astype(str)
        if 'Login Username' in df.columns:
            df['Login Username'] = df['Login Username'].astype(str)
        if 'Login Email' in df.columns:
            df['Login Email'] = df['Login Email'].astype(str)
        if 'Date' in df.columns:
            df['Date'] = df['Date'].apply(self._string2datetime)
        if 'Start Time' in df.columns:
            df['Start Time'] = df['Start Time'].apply(self._string2datetime)
        if 'Finish Time' in df.columns:
            df['Finish Time'] = df['Finish Time'].apply(self._string2datetime)
        if 'Duration (s)' in df.columns:
            df['Duration (s)'] = df['Start Time'] - df['Start Time']
            df['Duration (s)'] = df['Duration (s)'].apply(lambda x: x.total_seconds())
        if 'User' in df.columns:
            df['User'] = df['User'].astype(str)
        if 'Browser' in df.columns:
            df['Browser'] = df['Browser'].astype(str)
        if 'Device' in df.columns:
            df['Device'] = df['Device'].astype(str)
        if 'Referrer' in df.columns:
            df['Referrer'] = df['Referrer'].astype(str)
        return df

    def _sort_data(self, df: pd.DataFrame, ascending_bool: bool) -> pd.DataFrame:
        try:
            df = df.sort_values(by=['Reference #'], ascending=ascending_bool)
        except KeyError:
            try:
                df = df.sort_values(by=['Date'], ascending=ascending_bool)
            except KeyError:
                pass
        
        return df.reset_index(drop=True)

    def Process(self) -> pd.DataFrame:
        """Loads jsons in results list as dataframes and concats them."""
        results_list = []
        for results in [json['results'] for json in self.results]:
            results_list += results
        
        series_list = [self._process_row(json_row) for json_row in tqdm(results_list, desc="Processing results", 
                                                                        unit=' rows', 
                                                                        ncols=80, 
                                                                        dynamic_ncols=True, 
                                                                        leave=False)]

        dataframe = pd.DataFrame(series_list)
        combined_map = self.column_map
        combined_map.update(self.metadata_map)
        dataframe = self._reorder_columns(dataframe)
        dataframe.rename(columns=combined_map, inplace=True)
        dataframe = self._cast_dtypes(dataframe)
        if self.interface.params.last is not None:
            dataframe = self._sort_data(dataframe, False)
            until = int(self.interface.params.last)
            dataframe = dataframe.head(until)
        dataframe = self._sort_data(dataframe, self.sort_asc)
        return dataframe

    def _string2datetime(self, old_date: str) -> dt:
        """Converts ISO 8601 datetime string to datetime class."""
        try:
            new_date = dt.strptime(old_date, "%Y-%m-%dT%H:%M:%SZ") # ISO 8601 standard
            new_date = new_date + self.timezone_offset
            return new_date
        except TypeError:
            return old_date



from formsite_util.internal.interfaces import FormsiteParams, FormsiteInterface, FormsiteCredentials

# try:
#     from internal.interfaces import FormsiteParams, FormsiteInterface, FormsiteCredentials
# except ImportError:
#     from .internal.interfaces import FormsiteParams, FormsiteInterface, FormsiteCredentials


#!/usr/bin/env python3
"""
cli.py

CLI interface using argparse for formsite_util.core
"""
import csv # used in an eval
import sys
from time import perf_counter
import argparse
from formsite_util import FormsiteParams, FormsiteInterface, FormsiteCredentials
from formsite_util.internal.interfaces import __version__
# try:
#     from .internal.interfaces import FormsiteCredentials, FormsiteInterface, FormsiteParams, __version__
# except ImportError:
#     from internal.interfaces import FormsiteCredentials, FormsiteInterface, FormsiteParams, __version__

def gather_args() -> argparse.Namespace:
    """Gathers supported cli inputs."""
    parser = argparse.ArgumentParser(
        prog="formsite-util",
        description="Github of author: https://github.com/strny0/formsite-utility\n"
                    "This program performs an export of a specified formsite form with parameters.\n"
                    "A faster alternative to a manual export from the formsite website, that can be used for workflow automation.\n"
                    "Allows for the extraction of assets saved on formsite servers.",
        epilog="More info can be found at Formsite API v2 help page:    \n"
               "https://support.formsite.com/hc/en-us/articles/360000288594-API    \n"
               "You can find API related information of your specific form under: [Form Settings > Integrations > Formsite API] \n"
               "API response error codes table:\n"
               "| code | description                                 |\n"
               "| 401  | Authentication info is missing or invalid.  |\n"
               "| 403  | Forbidden.                                  |\n"
               "| 404  | Path or object not found.                   |\n"
               "| 422  | Invalid parameter.                          |\n"
               "| 429  | Too many requests or too busy.              |\n"
               "| 5xx  | Unexpected internal error.                  |\n\n"
               "formsite-util  Copyright (C) 2021  Jakub Strnad\n"
               "This program comes with ABSOLUTELY NO WARRANTY; for details see LICENSE.md\n"
               "This is free software, and you are welcome to redistribute it under certain conditions.",
        formatter_class=argparse.RawTextHelpFormatter
    )
    parser.add_argument('-V', '--version', action='version', version=f"%(prog)s {__version__}\n")
    g_auth = parser.add_argument_group('Authorization')
    g_params = parser.add_argument_group('Results Parameters')
    g_output = parser.add_argument_group('Output file')
    g_extract = parser.add_argument_group('Extract links')
    g_download = parser.add_argument_group('Download files')
    g_other = parser.add_argument_group('Other functions')
    g_debug = parser.add_argument_group('Debugging')
    g_nocreds = parser.add_mutually_exclusive_group(required=False)

    g_auth.add_argument('-t', '--token', type=str, default=None, required=True,
                        help="Your Formsite API token.")
    g_auth.add_argument('-s', '--server', type=str, default=None, required=True,
                        help="Your Formsite server. A part of the url. https://fsX.forms… <- the 'fsX' part."
                        "\nFor example 'fs22'.")
    g_auth.add_argument('-d', '--directory', type=str, default=None, required=True,
                        help="Your Formsite directory. Found under [Share > Links > Directory].")
    g_params.add_argument('-f', '--form', type=str, default=None,
                          help="Your Formsite form ID. Found under [Share > Links > Directory].")
    g_params.add_argument('--last', type=int, default=None,
                          help="Gets last x results."
                          "\nMust be an integer.")
    g_params.add_argument('--afterref', type=int, default=None,
                          help="Get results greater than a specified Reference #."
                          "\nMust be an integer.")
    g_params.add_argument('--beforeref', type=int, default=None,
                          help="Get results lesser than a specified Reference #."
                          "\nMust be an integer.")
    g_params.add_argument('--afterdate', type=str, default=None,
                          help="Get results after a specified date."
                          "\nMust be formatted as ISO 8601 UTC, YYYY-MM-DD, or YYYY-MM-DD HH:MM:SS."
                          "\nThis date is in your local timezone, unless specified otherwise.")
    g_params.add_argument('--beforedate', type=str, default=None,
                          help="Get results before a specified date."
                          "\nMust be formatted as ISO 8601 UTC, YYYY-MM-DD, or YYYY-MM-DD HH:MM:SS."
                          "\nThis date is in your local timezone, unless specified otherwise.")
    g_params.add_argument('--sort', choices=['asc', 'desc'], type=str, default="desc",
                          help="Determines how the output CSV will be sorted. Defaults to descending.")
    g_params.add_argument('--resultslabels', type=int, default=0,
                          help="Use specific results labels for your CSV headers."
                          "\nDefaults to 0 = first available set of results labels or default question labels.")
    g_params.add_argument('--resultsview', type=int, default=11,
                          help="Use specific results view for your CSV headers."
                          "\nDefaults to 11 = Items+Statistics. Other values currently not supported.")
    g_params.add_argument('-T', '--timezone', default='local',
                          help="Specify the timezone relative to which you want your results."
                          "\nBy default, results come relative to your local timezone."
                          "\nInput either name of the timezone from this list:"
                          "\nhttps://en.wikipedia.org/wiki/List_of_tz_database_time_zones"
                          "\nYou can specify a manual offset in format +02:00 or -0200 from your LOCAL time"
                          "\n[Examples: database names] (avoid using deprecated ones)"
                          "\n'America/Chicago'"
                          "\n'Europe/Paris'"
                          "\n'America/New_York'"
                          "\n'Asia/Bangkok'")
    g_params.add_argument('--delay', type=int, default=5, help='Delay in seconds between each API call.')
    g_output.add_argument('-o', '--output_file', nargs='?', default=False, const='default',
                          help="Specify output file name and location."
                          "\nDefaults to export_yyyymmdd_formID.csv in the folder of the script."
                          "\nSupported output formats are (.csv|.xlsx|.pkl|.pickle|.json|.parquet|.md|.txt)")
    g_output.add_argument('--encoding', type=str, default='utf-8',
                          help="Specify encoding of the output file (if output format supports it)."
                          "\nDefaults to 'utf-8'")
    g_output.add_argument('--separator', type=str, default=',',
                          help="Specify separator of the output file (if output format supports it)."
                          "\nDefaults to ',' (comma)")
    g_output.add_argument('--line_terminator', type=str, choices={"LF", "CR", "CRLF"}, default="LF",
                          help="Specify line terminator of the output file (if output format supports it)"
                          "\nDefaults to '\\n' (LF) (line feed)")
    g_output.add_argument('--quoting', type=str, choices={"MINIMAL", "ALL", "NONNUMERIC", "NONE"}, default="MINIMAL",
                          help="Specify quoting level of the output file (if output format supports it)."
                          "\nDefaults to 'MINIMAL'"
                          "\nMore info about the quoting levels: https://docs.python.org/3/library/csv.html")
    g_output.add_argument('--date_format', default="%Y-%m-%d %H:%M:%S",
                          help="Specify datetime format in output file using python strftime directives."
                          "\nDefaults to '%%Y-%%m-%%d %%H:%%M:%%S' which is yyyy-mm-dd HH:MM:SS"
                          "\nYou can find the possible format directives here:"
                          "\nhttps://docs.python.org/3/library/datetime.html#strftime-and-strptime-behavior")
    g_extract.add_argument('-x', '--extract_links', nargs='?', default=False, const='default',
                           help="Outputs a .txt file with all formsite files urls in specified form."
                           "\nYou can specify file name or location, for example:"
                           "\n'-x C:\\Users\\MyUsername\\Desktop\\download_links.txt'."
                           "\nIf you don't specify a location, it will default to the folder of the script.")
    g_download.add_argument('-D', '--download_links', nargs='?', default=False, const='default',
                            help="If you include this flag, all formsite links in the export will be downloaded to a folder."
                            "\nYou can specify location, for example `-D 'C:\\Users\\My Username\\Desktop\\downloads'"
                            "\nIf you don't specify a location, it will default to the folder of the script.")
    g_other.add_argument('-S', '--store_latest_ref', nargs='?', default=False, const='default',
                         help="If you enable this option, a text file `latest_ref.txt` will be created."
                         "\nThis file will only contain the highest reference number in the export."
                         "\nIf there are no results in your export, nothing will happen.")
    g_extract.add_argument('-X', '--links_regex', type=str, default='.+',
                           help="Keep only links that match the regex you provide."
                           "\nWon't do anything if -x or -d arguments are not provided."
                           "\nDefaults to '.+'. Example usage: '-X \\.json$' would only give you files that have .json extension.")
    g_download.add_argument('-c', '--concurrent_downloads', default=10, type=int,
                            help="You can specify the number of concurrent download tasks."
                            "\nMore for large numbers of small files, less for large files."
                            "\nDefault is 10")
    g_download.add_argument('-n', '--dont_overwrite_downloads', default=True, action="store_false",
                            help="Checks if files already exist in download directory based on filename, if yes, doesn't re-download them.")
    g_download.add_argument('-R', '--filename_regex', default='', type=str,
                            help="If you include this argument, filenames of the files you download from formsite "
                            "\nservers will remove all characters from their name that dont match the regex."
                            "\nExpecting an input of allowed characters, for example: -R '[^\\w\\_\\-]+'"
                            "\nAny files that would be overwritten as a result of the removal of characters will be appended with _1, _2, etc.")
    g_download.add_argument('--timeout', nargs=1, default=80, type=int,
                            help="Timeout in seconds for each individual file download."
                            "\nIf download time exceeds it, it will throw a timeout error and retry up until retries. Defaults to 80.")
    g_download.add_argument('--retries', nargs=1, default=1, type=int,
                            help="Number of times to retry downloading files if the download fails. Defaults to 1.")
    g_download.add_argument('--get_download_status', default=False, action='store_true',
                            help="If you enable this option, a text file with status for each downloaded link will be saved.")
    g_download.add_argument('--stripprefix', default=False, action='store_true',
                            help="If you enable this option, filenames of downloaded files will not have f-xxx-xxx- prefix.")
    g_nocreds.add_argument('-l', '--list_columns', action="store_true", default=False,
                           help="Outputs a table of column ID - column name mappings."
                           "\nRequires login info and form id. Overrides all other functionality of the program.")
    g_nocreds.add_argument('-L', '--list_forms', nargs='?', default=False, const='default',
                           help="By itself, prints all forms, their form ids and status. You can specify a file to save the data into."
                           "\nExample: '-L ./list_of_forms.csv' to output to file or '-L' by itself to print to console."
                           "\nRequires login info. Overrides all other functionality of the program.")
    g_debug.add_argument('--sort_list_by', default='name', choices={'name', 'form_id', 'resultsCount', 'filesSize'},
                           help="You may chose what to sort -L commands by in descending order. Defaults to name.")
    g_debug.add_argument('--disable_progressbars', action="store_true", default=False,
                         help="If you use this flag, program will not display progressbars to console.")
    g_debug.add_argument('-v', '--verbose', action="store_true", default=False,
                         help="If you use this flag, program will log progress in greater detail.")
    return parser.parse_known_args()[0]

def main():
    """formsite_util CLI entrypoint."""
    t_0 = perf_counter()
    arguments = gather_args()
    parameters = FormsiteParams(last=arguments.last,
                                afterref=arguments.afterref,
                                beforeref=arguments.beforeref,
                                afterdate=arguments.afterdate,
                                beforedate=arguments.beforedate,
                                resultslabels=arguments.resultslabels,
                                resultsview=arguments.resultsview,
                                timezone=arguments.timezone,
                                sort=arguments.sort)

    authorization = FormsiteCredentials(arguments.token,
                                        arguments.server,
                                        arguments.directory)

    interface = FormsiteInterface(arguments.form,
                                  authorization,
                                  params=parameters,
                                  verbose=arguments.verbose,
                                  display_progress=not arguments.disable_progressbars)

    if arguments.list_forms is not False:
        if arguments.list_forms == 'default':
            interface.ListAllForms(display=True, sort_by=arguments.sort_list_by)
        else:
            interface.ListAllForms(save2csv=arguments.list_forms)
            print(f'saved list of forms to {arguments.list_forms}')
        return 0

    if arguments.list_columns is not False:
        interface.ListColumns()
        return 0

    if arguments.output_file is not False:
        if interface.Data is None:
            interface.FetchResults()
        line_term = {"LF":"\n", "CR":"\r", "CRLF":"\r\n"}
        if arguments.output_file == 'default':
            default_filename = f'./export_{interface.form_id}_{interface.params.local_datetime.strftime("%Y-%m-%d--%H-%M-%S")}.csv'
        else:
            default_filename = arguments.output_file

        interface.WriteResults(default_filename,
                               date_format=arguments.date_format,
                               encoding=arguments.encoding,
                               separator=arguments.separator,
                               line_terminator=line_term.get(arguments.line_terminator),
                               quoting=getattr(csv, f'QUOTE_{arguments.quoting}'))
        print("export complete")

    if arguments.extract_links is not False:
        if interface.Data is None:
            interface.FetchResults()
        if arguments.extract_links == 'default':
            default_filename = f'./links_{interface.form_id}_{interface.params.local_datetime.strftime("%Y-%m-%d--%H-%M-%S")}.txt'
            interface.WriteLinks(
                default_filename, links_regex=arguments.links_regex)
        else:
            interface.WriteLinks(arguments.extract_links,
                                 links_regex=arguments.links_regex)
        print("links extracted")

    if arguments.download_links is not False:
        if interface.Data is None:
            interface.FetchResults()
        if arguments.download_links == 'default':
            default_folder = f'./download_{interface.form_id}_{interface.params.local_datetime.strftime("%Y-%m-%d--%H-%M-%S")}'
        else:
            default_folder = arguments.download_links

        interface.DownloadFiles(default_folder,
                                max_concurrent_downloads=arguments.concurrent_downloads,
                                links_regex=arguments.links_regex,
                                filename_regex=arguments.filename_regex,
                                overwrite_existing=arguments.dont_overwrite_downloads,
                                report_downloads=arguments.get_download_status,
                                timeout=arguments.timeout,
                                retries=arguments.retries,
                                strip_prefix=arguments.stripprefix)
        print("download complete")

    if arguments.store_latest_ref is not False:
        if interface.Data is None:
            interface.FetchResults()
        if arguments.store_latest_ref == 'default':
            default_filename = './latest_ref.txt'
        else:
            default_filename = arguments.store_latest_ref

        interface.WriteLatestRef(default_filename)
        print("latest reference saved")

    print(f'done in {(perf_counter() - t_0):0.2f} seconds!')
    return 0


if __name__ == '__main__':
    if sys.version_info[0] >= 3 and sys.version_info[1] > 5:
        sys.exit(main())
    else:
        raise Exception("Unsupported python version.")



"""
core.py

`FormsiteInterface` `FormsiteParams` and `FormsiteCredentials` classes are defined here.
Author: Jakub Strnad
Documentation: https://github.com/strny0/formsite-utility
"""
from __future__ import annotations
import csv
import asyncio
import json
from datetime import datetime as dt
from datetime import timedelta as td
from pathlib import Path
from collections import defaultdict
from dataclasses import dataclass
from typing import Any, Dict, Optional, Set, Union, Tuple, List
import re
import os
import pandas as pd
from pytz import UnknownTimeZoneError, timezone as pytztimezone
from aiohttp import request
from .downloader import _FormsiteDownloader
from .processing import _FormsiteProcessing
from .api import _FormsiteAPI
from tqdm import tqdm
__version__ = '1.3.9'

def _shift_param_date(date: Union[str, dt], timezone_offset: td) -> str:
    """Shifts input date in the string format/datetime by timedelta in timezone offset.
    The offset is additive, date + timezone_offset.

    Args:
        date (Union[str, dt]): String in formats: 'yyyy-mm-dd', 'yyyy-mm-dd HH:MM:SS' or 'yyyy-mm-ddTHH:MM:SSZ' or datetime. 
        timezone_offset (td): A timedelta value representing addition to 'date'.

    Raises:
        ValueError: Raised if string format is not recognized.

    Returns:
        str: A datetime string in 'yyyy-mm-ddTHH:MM:SSZ' format, shifted by timezone_offset amount.
    """    
    if isinstance(date, dt):
        date = date + timezone_offset
    else:
        try:
            try:
                date = dt.strptime(str(date), "%Y-%m-%dT%H:%M:%SZ")
                date = date + timezone_offset
            except:
                try:
                    date = dt.strptime(str(date), "%Y-%m-%d")
                    date = date + timezone_offset
                except:
                    try:
                        date = dt.strptime(str(date), "%Y-%m-%d %H:%M:%S")
                        date = date + timezone_offset
                    except:
                        raise
        except:
            raise ValueError("""invalid date format input for afterdate/beforedate, please use a datetime object or string in ISO 8601, yyyy-mm-dd or yyyy-mm-dd HH:MM:SS format""")

    return dt.strftime(date, "%Y-%m-%dT%H:%M:%SZ")

def _extract_timezone_from_str(timezone: str, timezone_re: str) -> td:
    """Parses input timezone: str -> timedelta

    Args:
        timezone (str): string in format ['+0200', '02:00', +02:00', '16:48', '-05:00', '-0600']
        timezone_re (str): regex encompassing above mentioned valid formats

    Returns:
        timedelta: timedelta offset
    """
    tz_offset = None
    if re.search(timezone_re, timezone) is not None:
        tz_str = timezone.replace(r'\'', '').replace(r'\"','')
        if ':' in tz_str:
            tz_tuple = tz_str.split(':', 1)
        else:
            t = (len(tz_str) - 2)
            tz_tuple = (tz_str[:t], tz_str[-2:])

        tz_offset = td(hours=int(tz_tuple[0]), seconds=int(tz_tuple[1])/60)
    return tz_offset

def _calculate_tz_offset(timezone: str) -> Tuple[td, dt]:
    """Calculates timezone offset relative to input TARGET timezone string and LOCAL timezone.

    Args:
        timezone (str): String in format Ex. 'America/Chicago' (tz_database_name) or offset like '0400', '+0400', '-7:00', '-14:00'

    Raises:
        UnknownTimeZoneError: If input is not a valid tz_databse name or offset.

    Returns:
        Tuple[timedelta, datetime]: tuple of offset_from_local time as a timedelta and local_time (datetime)
    """    
    local_date = dt.now()
    utc_date = dt.utcnow()
    utc_offset = local_date - utc_date
    if timezone == 'local':
        offset_from_local = td(seconds=0)
    else:
        tz_input_reges = r'(\+|\-|)([0-1]\d[0-5]\d|[0-1]\d\:[0-5]\d|\d\:[0-5]\d)'
        # (\+|\-|)         : '+' or '-' or ''
        # [0-1]\d[0-5]\d   : '0000'
        # [0-1]\d\:[0-5]\d : '00:00'
        # \d\:[0-5]\d      : '0:00'
        offset_from_local = _extract_timezone_from_str(timezone, tz_input_reges)
        if offset_from_local is None:
            if re.search(r'\w+/\w+', timezone) is not None:
                inp = pytztimezone(timezone).localize(local_date).strftime("%z")
                t = (len(inp) - 2)
                l_inp = (inp[:t], inp[-2:])
                inp_td = td(hours=int(l_inp[0]), seconds=int(l_inp[1])/60).total_seconds()
                offset_from_local = td(seconds=(inp_td - utc_offset.total_seconds()))
            else:
                raise UnknownTimeZoneError(timezone)

    return offset_from_local, local_date

def _sanitize_argument(argument: str, chars2remove: Dict[str,str]) -> str:
    """Performs a find and replace on 'argument' based on mapping in chars2remove.

    Args:
        argument (str): text to perform find and replace on
        chars2remove (List[str]): mapping of {find:replace}

    Returns:
        str: sanitized argument
    """
    for key, value in chars2remove.items():
        argument = str(argument).replace(key, value)
    return argument

def _confirm_arg_format(arg_value: str, arg_name: str, flag: str, example: str) -> str:
    """A boiler plate function to display a helpful error message.

    Args:
        arg_value (str): current value of the argument variable
        arg_name (str): exact name of the argument variable
        flag (str): flag used to invoke the argument in terminal
        example (str): correct example value

    Raises:
        ValueError: Raised upon entering the incorrect argumant/format.

    Returns:
        str: the sanitized (quote-less) argument back
    """
    quotes_map = {'\'':'', '\"':''}
    if not isinstance(arg_value, str):
        raise ValueError(f'invalid format for argument {arg_value}, {arg_name}, '
                        f'correct example: {flag} {example}')
    arg_value = _sanitize_argument(arg_value, quotes_map)
    return arg_value

def _validate_path(path: str) -> str:
    """Converts input path into POSIX format. Creates parent directories if necessary.

    Args:
        path (str): path to a file or a folder in any format

    Returns:
        str: path to a file or a folder in posix format
    """

    output_file = Path(path).resolve().absolute()
    if output_file.is_dir():
        os.makedirs(output_file.as_posix(), exist_ok=True)
    else:
        os.makedirs(output_file.parent.as_posix(), exist_ok=True)

    return output_file.as_posix()

@dataclass
class FormsiteParams:

    """Parameters class
    This class stores parameters for Formsite requests\n
    `afterref` gets only results greater than integer you provide\n
    `beforeref` gets only results less than integer you provide\n
    `afterdate` gets only results greater than input you provide, expects a `datetime object` or string in `ISO 8601`, `yyyy-mm-dd`, `yyyy-mm-dd HH:MM:SS`\n
    `beforedate` gets only results less than input you provide, expects a `datetime object` or string in `ISO 8601`, `yyyy-mm-dd`, `yyyy-mm-dd HH:MM:SS`\n
    `timezone` sets the timezone dates in results are relative to, also affects input dates. Expects either an offset string in format eg. `+06:00` or database name eg. `America/Chicago`\n
    `date_format` using python datetime directives specify what format you want your dates in your csv output file. Defaults to `%Y-%m-%d %H:%M:%S`\n
    `resultslabels` and `resultsview` More info on Formsite website or FS API of your specific form.\n
    `sort` ( "asc" | "desc" ) sorts results by reference number in ascending or descending order.
    """
    last: Optional[int] = None
    afterref: Optional[int] = None
    beforeref: Optional[int] = None
    afterdate: Optional[Union[str, dt]] = None
    beforedate: Optional[Union[str, dt]] = None
    timezone: Optional[str] = 'local'
    resultslabels: Optional[int] = None
    resultsview: Optional[int] = 11
    sort: Optional[str] = "desc"
    col_id_sort: Optional[int] = None
    search_equals: Optional[str] = None
    col_id_equals: Optional[int] = None
    search_contains: Optional[str] = None
    col_id_contains: Optional[int] = None
    search_begins: Optional[str] = None
    col_id_begins: Optional[int] = None
    search_ends: Optional[str] = None
    col_id_ends: Optional[int] = None
    search_method: Optional[str] = None

    def __post_init__(self):
        """Calls `_calculate_tz_offset` internal function."""
        self.timezone = 'local' if self.timezone is None else self.timezone
        self.timezone_offset, self.local_datetime = _calculate_tz_offset(self.timezone)

    def get_params_as_dict(self, single_page_limit: int = 500) -> dict:
        """Generates a parameters dictionary that is later passed to params= kw argument when making API calls.

        Args:
            single_page_limit (int, optional): Results per page limit, 500 is maximum amount. Defaults to 500.

        Returns:
            dict: params dict
        """
        results_params: dict[str, Union[str, int]] = dict()
        results_params['page'] = 1
        results_params['limit'] = single_page_limit
        if self.afterref is not None:
            results_params['after_id'] = self.afterref
        if self.beforeref is not None:
            results_params['before_id'] = self.beforeref
        if self.afterdate is not None:
            parsed_afterdate = _shift_param_date(self.afterdate, self.timezone_offset)
            results_params['after_date'] = parsed_afterdate
        if self.beforedate is not None:
            parsed_beforedate = _shift_param_date(self.beforedate, self.timezone_offset)
            results_params['before_date'] = parsed_beforedate
        # 11 = all items + statistics results view
        results_params['results_view'] = self.resultsview
        if self.col_id_sort is not None:
            results_params['sort_id'] = self.col_id_sort
        if self.col_id_equals is not None and self.search_equals is not None:
            results_params[f'search_equals[{self.col_id_equals}]'] = self.search_equals
        if self.col_id_contains is not None and self.search_contains is not None:
            results_params[f'search_contains[{self.col_id_contains}]'] = self.search_contains
        if self.col_id_begins is not None and self.search_begins is not None:
            results_params[f'search_begins[{self.col_id_begins}]'] = self.search_begins
        if self.col_id_ends is not None and self.search_ends is not None:
            results_params[f'search_ends[{self.col_id_ends}]'] = self.search_ends
        if self.search_method is not None:
            results_params['search_method'] = self.search_method
        return results_params

    def get_items(self) -> dict:
        """Returns a dict that gets parsed as parameters by aiohttp when making a request."""
        return {"results_labels": self.resultslabels}

@dataclass
class FormsiteCredentials:
    """Class representing formsite login information.

    Args:
        token (str): API access token

        server (str): formsite server, can be found in your formsite url at the beginning, https://fs_.(...).com

        directory (str): can also be found in your formsite URL, when accessing a specific form

    Returns:
        FormsiteCredentials instance
    """
    token: str
    server: str
    directory: str

    def __post_init__(self) -> None:
        """Confirms validity of input."""
        self.confirm_validity()

    def get_auth_header(self) -> dict:
        """Returns a dictionary sent as a header in the API request for authorization purposes."""
        return {"Authorization": f"bearer {self.token}", "Accept": "application/json"}

    def confirm_validity(self) -> None:
        """Checks if credentials input is in correct format."""
        self.token = _confirm_arg_format(self.token, 'token', '-t', 'token')
        self.server = _confirm_arg_format(self.server, 'server', '-s', 'fs1')
        self.directory = _confirm_arg_format(self.directory, 'directory', '-d', 'Wa37fh')

@dataclass
class FormsiteInterface:
    """A base class for interacting with the formsite API.\n
    Documentation: https://pypi.org/project/formsite-util/\n
    `self.Data` pandas dataframe storing your results.\n
    `self.Links` set storing all formsite links.\n
    Methods of interest:\n
    `FetchResults` stores results in self.Data of the instance of this class.\n
    `ReturnResults` returns a pandas dataframe with the results.\n
    `WriteResults` writes the dataframe to a file.\n
    `ExtractLinks` stores extracted links in self.Links of the instance of this class.\n
    `ReturnLinks` returns a touple with all links.\n
    `WriteLinks` writes them to a file.\n
    `ListAllForms` lists all forms on formsite, output them to console or save them to a file.\n
    `ListColumns` lists all columns and column IDs of a form you set the interface for.\n
    `DownloadFiles` downloads all files submitted to the form to a folder you specify.\n
    `WriteLatestRef` writes highest reference number in results to a file you specify.\n
    `display_progress = False` can be used to disable progressbars in console.
    """

    form_id: str
    auth: FormsiteCredentials
    params: FormsiteParams = FormsiteParams()
    verbose: bool = False
    Data: Optional[pd.DataFrame] = None
    Links: Optional[set] = None
    display_progress: bool = True
    api_call_delay: int = 5

    def __post_init__(self):
        """Initializes private variables.\n
        `url_base` is a base url in the format server.formsite.com/api/v2/directory\n
        `url_files` is url_base/files\n
        Also intializes HTTP headers for authorization and parameters.
        """
        self.url_base: str = f"https://{self.auth.server}.formsite.com/api/v2/{self.auth.directory}"
        self.url_files = f"https://{self.auth.server}.formsite.com/{self.auth.directory}/files/"
        self.auth_dict = self.auth.get_auth_header()
        self.params_dict = self.params.get_params_as_dict()
        self.items_dict = self.params.get_items()

    def __enter__(self):
        """Allows use of context managers."""
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Allows use of context managers."""
        del self

    def _perform_api_fetch(self) -> Tuple[dict, List[dict]]:
        """Entrypoint for performing API calls (asynchronously)."""
        api_handler = _FormsiteAPI(self, display_progress=self.display_progress, delay=self.api_call_delay)
        items, results = api_handler.Start()
        return items, results

    def _assemble_dataframe(self, items: dict, results: List[dict]) -> pd.DataFrame:
        """Returns a pandas dataframe from received API data."""
        if self.params.sort == 'desc':
            sort = False
        else:
            sort = True
        processing_handler = _FormsiteProcessing(items, results, self, sort_asc=sort, display_progress=self.display_progress)
        return processing_handler.Process()

    def FetchResults(self) -> None:
        """Fetches results from formsite API according to specified parameters.\n
        Updates the `self.Data` variable which stores the dataframe.
        """
        assert len(self.form_id) > 0, f"You must pass form id when instantiating FormsiteCredentials('form_id', login, params=...) you passed '{self.form_id}'"
        items, results = self._perform_api_fetch()
        self.Data = self._assemble_dataframe(items, results)

    def ReturnResults(self) -> pd.DataFrame:
        """Returns pandas dataframe of results."""
        if self.Data is None:
            self.FetchResults()
        return self.Data

    def _xtract(self, x: Any, pattern: re.Pattern, links_set: set = set(), links_filter_pattern: re.Pattern = None, pbar: tqdm = None):
        """Matches all formsite files links and appends them to links_set

        Args:
            x (Any): any element from self.Data
            pattern (re.Pattern): formsite files link pattern
            links_set (set, optional): output set of all links. Defaults to set().
            links_filter_pattern (re.Pattern, optional): regex to match links. Defaults to None.
        """
        _ = pbar.update(1) if pbar is not None else None
        try:
            for url_raw in pattern.findall(x):
                if '|' in url_raw:
                    for part in url_raw.split(' | '):
                        if links_filter_pattern.search(url_raw) is not None:
                            links_set.add(part)
                else:
                    if links_filter_pattern.search(url_raw) is not None:
                        links_set.add(url_raw)
        except Exception as ex:
            #print(ex)
            pass

    def ExtractLinks(self, links_filter_re: str = r'.+') -> None:
        """Stores a set of links in `self.Links` of files saved on formsite servers, that were submitted to the specified form."""
        if self.Data is None:
            self.FetchResults()
        links_re = fr'(https\:\/\/{self.auth.server}\.formsite\.com\/{self.auth.directory}\/files\/.*)'
        url_pattern = re.compile(links_re)
        filter_pattern = re.compile(links_filter_re)
        self.Links =set()
        if self.display_progress:
            pbar = tqdm(desc='Extracting download links', leave=False, unit=' cells')
        else:
            pbar = None
            
        _ = self.Data.applymap(lambda x: self._xtract(x, 
                                                      pattern=url_pattern, 
                                                      links_set=self.Links, 
                                                      links_filter_pattern=filter_pattern,
                                                      pbar=pbar))
        try:
            pbar.close()
        except AttributeError:
            pass

    def human_friendly_filesize(self, number: int) -> str:
        """Converts a number (filesize in bytes) to more readable filesize with units."""
        reductions = 0
        while number >= 1024:
            number = number / 1024
            reductions += 1
        unit = {0: '', 1:'K', 2:'M', 3:'G', 4:'T', 5:'P', 6:'E'}
        return f"{number:0.2f} {unit.get(reductions, None)}B"

    async def _list_all_forms(self) -> pd.DataFrame:
        url_forms = f"{self.url_base}/forms"
        async with request("GET", url_forms, headers=self.auth_dict) as response:
            response.raise_for_status()
            content = await response.content.read()
            all_forms_json = json.loads(content.decode('utf-8'))['forms']
            # un-nest the stats object
            for row in all_forms_json:
                for val in row["stats"]:
                    row[val] = row['stats'][val]
            for row in all_forms_json:
                for val in row["publish"]:
                    row[val] = row['publish'][val]
            forms_df = pd.DataFrame(all_forms_json,columns=['name', 'state', 'directory', 'resultsCount', 'filesSize', 'embed_code', 'link'])
            forms_df['form_id'] = forms_df.pop('directory')
            return forms_df

    def ListAllForms(self,
                     sort_by: str = 'name',
                     display: bool = False,
                     save2csv: Union[str, bool] = False) -> pd.DataFrame:
        """Prints name, id, results count, filesize and status of all forms into console or csv.
        You can sort in descending order by `name` `form_id` `resultsCount` `filesSize`.

        Args:
            sort_by (str, optional): One of {'name', 'form_id', 'resultsCount', 'filesSize'}. Defaults to 'name'.
            display (bool, optional): Print to stdout. Defaults to False.
            save2csv (Union[str, bool], optional): Path to output csv file. Defaults to False.
        
        Returns:
            pd.DataFrame: Dataframe with all relevant forms data.    
        """
        forms_df = asyncio.get_event_loop().run_until_complete(self._list_all_forms())
        if display:
            pd.set_option('display.max_rows', None)
            pd.set_option('display.max_columns', None)
            pd.set_option('display.width', None)
            pd.set_option('display.max_colwidth', 42) # ensures width < 80 cols
            forms_df = forms_df[['name','resultsCount','filesSize','form_id']]
            forms_df.sort_values(by=[sort_by], inplace=True, ascending=False, ignore_index=True)
            forms_df = forms_df.append({'name':'Total:', 'resultsCount':forms_df['resultsCount'].sum(), 'filesSize':forms_df['filesSize'].sum(),'form_id':f'{forms_df.shape[0]} forms'}, ignore_index=True)
            forms_df['filesSize'] = forms_df['filesSize'].apply(lambda x: self.human_friendly_filesize(int(x)))
            forms_df.set_index('name',inplace=True)
            print(forms_df)

        if save2csv is not False:
            forms_df.sort_values(by=[sort_by], inplace=True, ascending=False)
            forms_df.set_index('name',inplace=True)
            output_file = _validate_path(str(save2csv))
            forms_df.to_csv(output_file, encoding='utf-8')
            
        return forms_df

    def ReturnLinks(self, links_regex: str = r'.+') -> Set[str]:
        """Returns a set of urls of files saved on formsite servers."""
        if self.Links is None or links_regex != r'.+':
            self.ExtractLinks(links_filter_re=links_regex)
        return self.Links

    def WriteLinks(self,
                   destination_path: str,
                   links_regex: str = r'.+',
                   sort_descending: bool = True) -> None:
        """Writes links extracted with `self.ExtractLinks()` to a .txt file.

        Args:
            destination_path (str): path to output file
            links_regex (str, optional): Include only links that match target regex. Defaults to r'.+'.
            sort_descending (bool, optional): Defaults to True.
        """
        if self.Links is None or links_regex != r'.+':
            self.ExtractLinks(links_filter_re=links_regex)
        output_file = _validate_path(destination_path)
        sorted_links = [link + '\n' for link in self.Links]
        sorted_links.sort(reverse=sort_descending)
        with open(output_file, 'w') as writer:
            writer.writelines(sorted_links)

    def DownloadFiles(self,
                      download_folder: str,
                      max_concurrent_downloads: int = 10,
                      links_regex: str = r'.+',
                      filename_regex: str = r'',
                      overwrite_existing: bool = True,
                      report_downloads: bool = False,
                      timeout: float = 80,
                      retries: int = 1,
                      strip_prefix: bool = False) -> None:
        """Downloads files saved on formsite servers, that were submitted to the specified form.
        Please customize `max_concurrent_downloads` to your specific use case.

        Args:
            download_folder (str): Path to target download folder
            max_concurrent_downloads (int, optional): Defaults to 10.
            links_regex (str, optional): Example: r'.+\.jpg$' would get all files that end with .jpg. Defaults to r'.+'.
            filename_regex (str, optional): Removes characters that don't match regex from remote-filename. Defaults to r''.
            overwrite_existing (bool, optional): Whether or not to overwrite existing files. Defaults to True.
            report_downloads (bool, optional): Generates a report .txt file. Defaults to False.
            timeout (float, optional): In seconds, specify how long to wait for connection/download. Defaults to 80.
            retries (int, optional): In case of failed download or a timeout error, how many times to retry download. Defaults to 1.
            strip_prefix (bool, optional): If True, strips f-xxx-xxx prefix. Defaults to False.
        
        Raises:
            AssertationError: if len(self.Links) < 1, ie. there is nothing to download.
        """
        if self.Links is None:
            self.ExtractLinks(links_filter_re=links_regex)
        assert len(self.Links) > 0, f"There are no files to be downloaded for form {self.form_id}"

        download_folder = _validate_path(download_folder)
        download_handler = _FormsiteDownloader(download_folder,
                                               self.Links,
                                               max_concurrent_downloads,
                                               overwrite_existing=overwrite_existing,
                                               filename_regex=filename_regex,
                                               report_downloads=report_downloads,
                                               timeout=timeout,
                                               retries=retries,
                                               display_progress=self.display_progress,
                                               strip_prefix=strip_prefix)
        asyncio.get_event_loop().run_until_complete(download_handler.Start())

    def ListColumns(self) -> None:
        """Prints list of columns (items, usercontrols) and their respective formsite IDs."""
        api_handler = _FormsiteAPI(self)
        api_handler.check_pages = False
        items = asyncio.get_event_loop().run_until_complete(
            api_handler.Start(only_items=True))
        items = pd.DataFrame(json.loads(items)['items'], columns=['id', 'label', 'position'])
        items = items.set_index('id')
        pd.set_option('display.max_rows', None)
        print(items)
        print('----------------')
        print(f"Results labels: {self.params.resultslabels}")
        print(f"Results view: {self.params.resultsview}")

    def WriteResults(self, destination_path: str,
                     encoding: str = "utf-8",
                     line_terminator: str = '\n',
                     separator: str = ",",
                     date_format: str = "%Y-%m-%d %H:%M:%S",
                     quoting: int = csv.QUOTE_MINIMAL) -> None:
        """Writes `self.Data` to a file based on provided extension.
        Supported output formats are (`.csv`|`.xlsx`|`.pkl`|`.pickle`|`.json`|`.parquet`|`.md`|`.txt`)

        Args:
            destination_path (str): path to output file
            encoding (str, optional): Defaults to "utf-8".
            line_terminator (str, optional): Defaults to '\n'.
            separator (str, optional): Defaults to ",".
            date_format (str, optional): Defaults to "%Y-%m-%d %H:%M:%S".
            quoting (int, optional): Pass values from csv.QUOTE_ enum. Defaults to csv.QUOTE_MINIMAL.
        """
        if self.Data is None:
            self.FetchResults()
        output_file = _validate_path(destination_path)
        if re.search('.+\\.txt$', output_file) is not None:
            self.Data.to_string(output_file, encoding=encoding, index=False)
        elif re.search('.+\\.pkl$', output_file) is not None or re.search('.+\\.pickle$', output_file) is not None:
            self.Data.to_pickle(output_file)
        elif re.search('.+\\.parquet$', output_file) is not None:
            self.Data.to_parquet(output_file, index=False)
        elif re.search('.+\\.md$', output_file) is not None:
            self.Data.to_markdown(output_file, index=False)
        elif re.search('.+\\.json$', output_file) is not None:
            try:
                self.Data.to_json(output_file, orient='records', date_format='iso')
            except ValueError:
                renamer: dict[Any, Any] = defaultdict()
                for column_name in self.Data.columns[self.Data.columns.duplicated(keep=False)].tolist():
                    if column_name not in renamer:
                        renamer[column_name] = [column_name+'_0']
                    else:
                        renamer[column_name].append(
                            column_name + '_'+str(len(renamer[column_name])))
                self.Data.rename(columns=lambda column_name: renamer[column_name].pop(
                    0) if column_name in renamer else column_name).to_json(output_file, orient='records', date_format='iso')
        elif re.search('.+\\.xlsx$', output_file) is not None:
            print('Writing to excel (this can be slow for large files!)')
            self.Data.to_excel(output_file, index=False,
                               engine='openpyxl', freeze_panes=(1, 1))
        else:
            self.Data.to_csv(output_file, index=False, chunksize=1024, encoding=encoding,
                             date_format=date_format, line_terminator=line_terminator, sep=separator, quoting=quoting)

    def WriteLatestRef(self, destination_path: str) -> None:
        """Writes `max(self.Data['Reference #])` to a file."""
        if self.Data is None:
            self.FetchResults()

        output_file = _validate_path(destination_path)
        latest_ref = max(self.Data['Reference #'])
        with open(output_file, 'w') as writer:
            writer.write(str(latest_ref))



"""
api.py

This module contains a class that handles API requests to Formsite API.
"""
from __future__ import annotations
import json
from typing import List, Tuple, Any
from dataclasses import dataclass
import time
import requests
from requests.models import HTTPError
from tqdm import tqdm

@dataclass
class _FormsiteAPI:
    """Handles retreival of results from your formsite form. Invoked with `self.Start()` method.
    \nKeywords:
    \nResults: Data from your formsite form (the rows).
    \nItems:   Names of user-defined columns in your formsite form (the header row).
    \nInputs:
    \ninterface: instance of FormsiteInterface class from core.py
    \nsave_results_jsons: whether to write received results to json files
    \nsave_items_json: whether to write received itmes to json files"""
    interface: Any
    save_results_jsons: bool = False
    save_items_json: bool = False
    display_progress:bool = True
    delay: float = 5.0
    long_delay: float = 15.0

    def __post_init__(self):
        """Generates post_init internal variables."""
        self.params_dict: dict = self.interface.params_dict
        self.items_dict: dict = self.interface.items_dict
        self.auth_dict: dict = self.interface.auth_dict
        self.form_id: str = self.interface.form_id
        self.results_url: str = f'{self.interface.url_base}/forms/{self.interface.form_id}/results'
        self.items_url: str = f'{self.interface.url_base}/forms/{self.interface.form_id}/items'
        self.total_pages: int = 1
        self.check_pages: bool = True
        self.pbar = tqdm(desc='Starting API requests', total=2, unit='call', leave=False, dynamic_ncols=True, ncols=80) if self.display_progress else None
        self.session = None
        #self.save_results_jsons = Path(f'./{self.form_id}_tmp_results/').resolve().absolute().as_posix()
        #self.save_items_json = Path(f'./{self.form_id}_tmp_items/').resolve().absolute().as_posix()
        #os.makedirs(self.save_results_jsons, exist_ok=True)
        #os.makedirs(self.save_items_json, exist_ok=True)

    def Start(self, only_items: bool = False) -> Tuple[dict, List[dict]]:
        """Performs all API calls to formsite servers asynchronously"""
        items, results = (None, [])
        with requests.session() as self.session:
            self.session.headers.update(self.auth_dict)
            try:
                results = self.get_results() if not only_items else None
            except Exception as err:
                print(err)
                try:
                    del self.params_dict['after_id']
                except KeyError as kerr: 
                    pass
                try:
                    self.params_dict['before_id'] = int(results[-1]['results'][-1]['id'])
                except Exception as err2:
                    print(err2)
                    print('\n\n\n\n\nThis error was hard to catch, please report it if you see it.')
                    quit()
                results += self.get_results() if not only_items else None
            items = self.fetch_items()
            self._update_pbar_progress()
            self._update_pbar_desc(desc="API calls complete")
            try:
                self.pbar.close()
            except AttributeError:
                pass
        #shutil.rmtree(path=self.save_results_jsons, ignore_errors=True)
        #shutil.rmtree(path=self.save_items_json, ignore_errors=True)
        return items, results
    
    def get_results(self) -> List[dict]:
        """Fetches all results that match input parameters from a form.

        Returns:
            List[dict]: List of pages of results.
        """
        results = [self.fetch_results(self.params_dict, 1)]
        if self.total_pages > 1:
            for page in range(2, self.total_pages+1):
                results.append(self.fetch_results(self.params_dict, page))
                if isinstance(self.interface.params.last, int):
                    self.pbar.total = int(self.interface.params.last % 500)
                    if (page * 500) >= self.interface.params.last:
                        break
        return results
                
    def fetch_results(self, params: dict, page: int) -> str:
        """Handles fetching and writing (if selected) of results json."""
        self._update_pbar_desc(desc=f"Fetching results ({(page-1)*500}-{page*500})")
        content = self.fetch_content(self.results_url, params, page=page)
        # if len(content['results']) < 1:
        #     shutil.rmtree(path=self.save_results_jsons, ignore_errors=True)
        #     shutil.rmtree(path=self.save_items_json, ignore_errors=True)
        #     raise ValueError(f"[{self.form_id}] No results within specified parameters: {self.params_dict}")
        # max_ref = content['results'][0]['id']
        # min_ref = content['results'][-1]['id']
        # self.write_content(f'{self.save_results_jsons}/results_{self.form_id}_{min_ref}_{max_ref}.json', content)
        return content

    def write_content(self, filename: str, content: dict) -> None:
        """Handles writing of any string (json)."""
        with open(filename, 'w') as writer:
            json.dump(content, writer)

    def fetch_content(self, url: str, params: dict, page: int = None) -> dict:
        """Base method for interacting with the formsite api with aiohttp GET request. Returns content of the response."""  
        if page is not None:
            params['page'] = page
            if page > 3:
                self._update_pbar_desc(desc=f"Delay [{self.long_delay:0.0f} s] ({(page-1)*500}-{page*500})")
                _ = time.sleep(abs(self.long_delay-self.delay))
            else:
                self._update_pbar_desc(desc=f"Delay [{self.delay:0.0f} s] ({(page-1)*500}-{page*500})")
        time.sleep(self.delay)
        self._update_pbar_desc(desc="Fetching results")
        with self.session.get(url, params=params) as response:
            content = response.json()
            if response.status_code != 200:
                self._update_pbar_desc(desc=f"Error: [{response.status_code}]")
                if response.status_code == 429:
                    self._update_pbar_desc(desc="Reached rate limit, waiting 60 seconds")
                    time.sleep(60)
                    self._update_pbar_desc(desc="Fetching results")
                    content = self.fetch_content(url, params, page=page)
                else:
                    response.raise_for_status()
                    raise HTTPError(f"\r[{response.status_code}] {response.content}")
            if self.check_pages:
                self.total_pages = int(response.headers['Pagination-Page-Last'])
                self.check_pages = False
                try:
                    self.pbar.total = self.total_pages
                except AttributeError as ex:
                    pass
        self._update_pbar_progress()
        return content


    def fetch_items(self) -> str:
        """Handles fetching and writing (if selected) of items json."""
        self._update_pbar_desc(desc="Fetching headers")
        content = self.fetch_content(self.items_url, self.items_dict)
        # self.write_content(f'{self.save_items_json}/items_{self.form_id}.json', content)
        return content

    def _update_pbar_progress(self) -> None:
        if isinstance(self.pbar, tqdm):
            self.pbar.update(1)

    def _update_pbar_desc(self, desc: str) -> None:
        if isinstance(self.pbar, tqdm):
            self.pbar.set_description(desc=desc, refresh=True)



"""Module doscstring to make pylint STFU."""

import setuptools
from formsite_util.internal.interfaces import __version__

with open('README.md', 'r', encoding='utf-8') as reader:
    long_description = reader.read()

packages = setuptools.find_packages()

setuptools.setup(
    name='formsite-util',
    version=__version__,
    author='Jakub Strnad',
    author_email='jakub.strnad@protonmail.com',
    description='A simple Formsite API python script to get results and download files from your forms.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=['asyncio', 'aiohttp','requests','colorama', 'dataclasses', 'typing',
                      'pandas', 'python_dateutil', 'pytz', 'regex', 'tqdm', 'prompt_toolkit'],
    keywords=['python', 'formsite', 'fs', 'api', 'automation', 'download', 'form', 'python3',
              'utility', 'util', 'system', 'rest', 'integration', 'links', 'url', 'urls',
              'cli', 'gui', 'py'],
    url='https://github.com/strny0/formsite-utility/',
    project_urls={
        'Bug Tracker': 'https://github.com/strny0/formsite-utility/issues',
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Development Status :: 5 - Production/Stable',
        'Topic :: Utilities',
        'Typing :: Typed',
        'Environment :: Console',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)'
    ],
    packages=packages,
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            'getform=formsite_util.terminal_cli:main',
        ]
    }
)
